package edu.washington.cs.radishsim;

import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

interface CacheCallbacks<Line extends CoherentLine> {
   /**
    * Called whenever a line needs to be evicted.
    * 
    * @param set
    *           the set from which we need to evict something; the list is ordered from MRU (front)
    *           to LRU (back).
    * @param level
    *           the level of the cache where the eviction is happening
    * @return the line to evict
    */
   Line eviction(final Line incoming, final Deque<Line> set, CacheLevel level);
}

class CacheConfiguration<Line extends CoherentLine> {
   public int        cacheSize;
   public int        assoc;
   public int        lineSize;
   public CacheLevel level;
}

enum CacheLevel {
   L1, L2, L3, MEMORY
};

/** The response from sending a request to the memory hierarchy */
class MemoryResponse<Line extends CoherentLine> {
   /** The level of the cache hierarchy where the hit occurred */
   public CacheLevel whereHit;
   /** The line containing the requested address. */
   public Line       lineHit;

   @Override
   public String toString() {
      return whereHit.toString() + " " + String.valueOf(lineHit);
   }
}

public class HierarchicalCache<Line extends CoherentLine> {

   protected CacheLevel           levelInHierarchy = CacheLevel.L1;

   /** associativity */
   protected int                  assoc;

   /** log_2(line size) */
   protected short                lineOffsetBits;
   /** log_2(num sets) */
   protected short                indexBits;
   /** mask used to clear out the tag bits */
   protected long                 indexMask;

   protected List<Deque<Line>>    sets;
   protected CacheCallbacks<Line> callbacks;
   protected LineFactory<Line>    lineFactory;

   /**
    * The next higher-level cache in the hierarchy. Can be shared by multiple lower-level caches.
    */
   public HierarchicalCache<Line> nextCache;

   /** Return the cache index for the given address */
   protected int index(long address) {
      return (int) ((address >> lineOffsetBits) & indexMask);
   }

   /**
    * Put the incoming line into this cache.
    * 
    * @param incoming
    *           the line being evicted from a lower-level cache
    */
   protected void evictedFromLowerCache(final Line incoming) {
      if ( incoming.invalid() ) {
         // ignore the eviction of invalid lines
         return;
      }
      LineAddress la = incoming.lineAddress();

      Deque<Line> set = sets.get(index(la.get()));

      Line toEvict = callbacks.eviction(incoming, set, levelInHierarchy);
      boolean found = set.remove(toEvict);
      assert found;

      // NB: add the incoming line *after* the eviction handler runs
      set.addFirst(incoming);

      if ( nextCache != null ) {
         nextCache.evictedFromLowerCache(toEvict);
      }
   }

   /**
    * Construct a component of the cache hierarchy. NB: the components must be constructed in
    * reverse level order, from highest/last-level -> lowest-level.
    * 
    * @param thisConfig
    *           geometry and event handlers for this cache
    * @param nextCache
    *           the next higher cache in the hierarchy
    * @param factory
    *           factory for making Line objects
    */
   public HierarchicalCache(CacheConfiguration<Line> thisConfig, CacheCallbacks<Line> handler,
                            HierarchicalCache<Line> nextCache, LineFactory<Line> factory) {
      this.assoc = thisConfig.assoc;
      this.callbacks = handler;
      this.nextCache = nextCache;
      this.lineFactory = factory;
      this.levelInHierarchy = thisConfig.level;

      // construct the cache itself

      assert BitTwiddle.isPowerOf2(thisConfig.lineSize);
      assert BitTwiddle.isPowerOf2(thisConfig.assoc);
      assert BitTwiddle.isPowerOf2(thisConfig.cacheSize);

      int numSets = thisConfig.cacheSize / (thisConfig.lineSize * thisConfig.assoc);
      assert BitTwiddle.isPowerOf2(numSets);
      assert numSets * thisConfig.lineSize * thisConfig.assoc == thisConfig.cacheSize;
      indexMask = numSets - 1;
      indexBits = (short) BitTwiddle.floorLog2(numSets);
      lineOffsetBits = (short) BitTwiddle.floorLog2(thisConfig.lineSize);

      sets = new Vector<Deque<Line>>(numSets);

      for ( int i = 0; i < numSets; i++ ) {
         Deque<Line> set = new LinkedList<Line>();
         for ( int j = 0; j < thisConfig.assoc; j++ ) {
            Line line = lineFactory.create();
            set.add(line);
         }
         sets.add(set);
      }

   } // end ctor

   /** Make a request, creating compact lines on fills. */
   public MemoryResponse<Line> request(final ByteAddress address) {
      return request(address, RadishLine.Compacted.COMPACT);
   }

   /**
    * Access the line containing the given address, bringing it into the L1 cache if it was not
    * already present there.
    * 
    * While this function is recursing, each intermediate lines returned (if not already in the L1)
    * has been removed from its containing cache, and is ready to be inserted into the L1.
    * 
    * @param address
    *           the memory address being accessed
    * @param c
    *           whether to create compact or expanded lines, if lines need to be filled
    * @return an AccessReturn object that 1) says where in the memory hierarchy the requested line
    *         was found and 2) holds a reference to the new line (now in the L1)
    */
   public MemoryResponse<Line> request(final ByteAddress address, RadishLine.Compacted c) {
      Deque<Line> set = sets.get(index(address.get()));
      assert (set.size() == assoc);

      MemoryResponse<Line> ret = new MemoryResponse<Line>();

      // search this cache
      for ( Iterator<Line> it = set.iterator(); it.hasNext(); ) {
         Line line = it.next();
         if ( line.valid() && line.contains(address) ) {
            // hit! move this line to the mru spot
            if ( levelInHierarchy == CacheLevel.L1 ) {
               if ( line != set.getFirst() ) {
                  it.remove();
                  set.addFirst(line);
               }
               ret.lineHit = line;
               assert line == set.getFirst();

            } else { // L2+ cache
               it.remove();
               set.addLast(lineFactory.create());
               ret.lineHit = line; // allow line to be pulled into L1
            }

            ret.whereHit = this.levelInHierarchy;
            return ret;
         }
      }

      // if we made it here, we missed in this cache

      // check higher-level caches
      if ( nextCache != null ) {
         ret = nextCache.request(address, c);
         // response should have come from deeper in the hierarchy
         assert ret.whereHit.compareTo(levelInHierarchy) > 0;

      } else { // missed in the LLC
         ret.whereHit = CacheLevel.MEMORY;
      }

      // only L1 caches have to deal with evictions here;
      // other levels are handled via evictedFromLowerCache()
      if ( levelInHierarchy != CacheLevel.L1 ) return ret;

      switch ( ret.whereHit ) {
      case MEMORY:
         // bring line from memory straight into L1
         ret.lineHit = lineFactory.create(address.lineAddress(), c);
         break;
      case L1:
      case L2:
      case L3:
         // bring in hit line from higher-level cache
         assert ret.lineHit != null;
         break;
      default:
         assert false;
      }

      // evict a line (possibly to next-level cache)
      Line toEvict = callbacks.eviction(ret.lineHit, set, levelInHierarchy);
      boolean found = set.remove(toEvict);
      assert found;

      // NB: only push in the incoming line *after* we've evicted something
      set.addFirst(ret.lineHit);

      if ( nextCache != null ) {
         nextCache.evictedFromLowerCache(toEvict);
      }

      return ret;

   } // end request()

   /**
    * Lookup the given address in the memory hierarchy.
    * 
    * @param address
    *           the memory address being accessed
    * @param reorderSet
    *           If true and the line is found in this cache, move the line to MRU position. If
    *           false, do not modify the set ordering.
    * @return an AccessReturn object that 1) says where in the memory hierarchy the requested line
    *         was found and 2) holds a reference to the line (which may reside in any level of the
    *         hierarchy). If the address is not cached, then the line reference is null.
    */
   public MemoryResponse<Line> search(final DataByteAddress address, final boolean reorderSet) {
      return __search(address, reorderSet);
   }

   /** Search for this metadata address in the cache hierarchy, taking line ownership into account. */
   public MemoryResponse<Line> ownedMdSearch(final MetadataByteAddress address,
                                             final boolean reorderSet) {
      return __search(address, reorderSet);
   }

   private MemoryResponse<Line> __search(final ByteAddress address, final boolean reorderSet) {
      MemoryResponse<Line> ret = new MemoryResponse<Line>();
      final Deque<Line> set = sets.get(index(address.get()));

      // search this cache
      for ( Line l : set ) {

         if ( l.valid() && l.contains(address) ) {
            // hit!
            ret.lineHit = l;
            ret.whereHit = levelInHierarchy;

            if ( reorderSet ) {
               set.remove(l);
               set.addFirst(l);
            }

            return ret;
         }
      }

      // at this point, we missed in this cache

      if ( nextCache != null ) { return nextCache.__search(address, reorderSet); }

      // we're the last cache in the hierarchy
      // line is unmodified
      ret.whereHit = CacheLevel.MEMORY;
      ret.lineHit = null;
      return ret;

   } // end search()

   public interface LineVisitor<Line> {
      public void visit(Line l);
   }

   /**
    * Calls the given visitor function once on each line in this cache. Lines are traversed in no
    * particular order.
    */
   public void visitAllLines(LineVisitor<Line> lv) {
      for ( Deque<Line> set : sets ) {
         for ( Line l : set ) {
            lv.visit(l);
         }
      }
   }

   /** Verify that each line is indexed into the proper set. */
   public void verifyIndices() {
      for ( int i = 0; i < sets.size(); i++ ) {
         Deque<Line> set = sets.get(i);
         for ( Line l : set ) {
            if ( l.lineAddress() != null ) {
               assert index(l.lineAddress().get()) == i;
            }
         }
      }
   }
   
   @Override
   public String toString() {
      StringBuilder s = new StringBuilder("cache=" + this.levelInHierarchy
                                          + System.getProperty("line.separator"));
      for ( Deque<Line> set : sets ) {
         for ( Line l : set ) {
            s.append(l.toString() + "; ");
         }
         s.append(System.getProperty("line.separator"));
      }

      if ( nextCache != null ) s.append(nextCache.toString());
      return s.toString();
   }

} // end class HierarchicalCache
