package edu.washington.cs.radishsim;

/**
 * These types would be encoded as "variant types" in other languages, but have to be distinct
 * classes in Java. Basically, we want to maintain the distinctions of different kinds of numbers,
 * instead of treating them all as ints/longs.
 */

/** Tracks CPU id's */
class CpuId {
   private final short cpuid;

   public CpuId(int id) {
      cpuid = (short) id;
   }

   public short get() {
      return cpuid;
   }

   @Override
   public boolean equals(Object o) {
      if ( o instanceof CpuId ) {
         CpuId other = (CpuId) o;
         return this.cpuid == other.cpuid;
      }
      return false;
   }

   @Override
   public int hashCode() {
      return this.cpuid;
   }

   @Override
   public String toString() {
      return "p" + this.cpuid;
   }
}

/** Tracks thread id's */
class ThreadId {
   private final short tid;

   public ThreadId(int id) {
      tid = (short) id;
   }

   public short get() {
      return tid;
   }

   public boolean equals(short a) {
      return a == tid;
   }

   @Override
   public boolean equals(Object o) {
      if ( o instanceof ThreadId ) {
         ThreadId other = (ThreadId) o;
         return this.tid == other.tid;
      }
      return false;
   }

   @Override
   public int hashCode() {
      return this.tid;
   }

   @Override
   public String toString() {
      return "t" + this.tid;
   }
}

/** The fully-qualified address of a particular byte in memory */
abstract class ByteAddress {
   protected long addr;

   public ByteAddress(long a) {
      this.addr = a;
   }

   public long get() {
      return this.addr;
   }

   public void incr() {
      this.addr++;
   }

   public void incr(long a) {
      this.addr += a;
   }

   abstract public LineAddress lineAddress();

   public short lineOffset() {
      return (short) (this.addr & MemorySystemConstants.LINE_OFFSET_MASK());
   }

   @Override
   public String toString() {
      return "ba:" + String.format("%d", this.addr);
   }
}

/** A cache-line-aligned address */
abstract class LineAddress {
   protected final long lineAddr;

   public LineAddress(long la) {
      assert (la & (~MemorySystemConstants.LINE_OFFSET_MASK())) == la;
      this.lineAddr = la;
   }

   public long get() {
      return this.lineAddr;
   }

   @Override
   public boolean equals(Object o) {
      if ( o instanceof LineAddress ) {
         LineAddress other = (LineAddress) o;
         return this.lineAddr == other.lineAddr;
      }
      return false;
   }

   @Override
   public int hashCode() {
      return (int) this.lineAddr;
   }

   @Override
   public String toString() {
      return "la:" + String.format("%d", this.lineAddr);
   }
}

/** The address of a regular data byte. */
class DataByteAddress extends ByteAddress {
   public DataByteAddress(long a) {
      super(a);
   }

   public LineAddress lineAddress() {
      return new DataLineAddress(this.addr & (~MemorySystemConstants.LINE_OFFSET_MASK()));
   }

   @Override
   public String toString() {
      return "dba:" + String.format("%d", this.addr);
   }
}

/** The line address of a regular data byte. */
class DataLineAddress extends LineAddress {
   public DataLineAddress(long la) {
      super(la);
   }

   @Override
   public boolean equals(Object o) {
      if ( o instanceof DataLineAddress ) {
         DataLineAddress other = (DataLineAddress) o;
         return this.lineAddr == other.lineAddr;
      }
      return false;
   }

   @Override
   public int hashCode() {
      return (int) this.lineAddr;
   }

   @Override
   public String toString() {
      return "dla:" + String.format("%d", this.lineAddr);
   }
}

/** The line address of a metadata byte. */
class MetadataLineAddress extends LineAddress {
   protected CpuId owner;

   /** NB: la does not undergo translation */
   public MetadataLineAddress(long la, CpuId own) {
      super(la);
      assert own != null;
      this.owner = own;
   }

   public CpuId owner() {
      return this.owner;
   }

   @Override
   public boolean equals(Object o) {
      if ( o instanceof MetadataLineAddress ) {
         MetadataLineAddress other = (MetadataLineAddress) o;
         return this.lineAddr == other.lineAddr && this.owner.equals(other.owner);
      }
      return false;
   }

   @Override
   public int hashCode() {
      return (int) this.lineAddr ^ this.owner.hashCode();
   }

   @Override
   public String toString() {
      return "mla:" + String.format("%d", this.lineAddr) + "," + this.owner.toString();
   }
}

/** The line address of a metadata byte, without an owner. */
class UnownedMetadataLineAddress extends LineAddress {

   /**
    * @param la
    *           does not undergo translation
    */
   public UnownedMetadataLineAddress(long la) {
      super(la);
   }

   @Override
   public boolean equals(Object o) {
      if ( o instanceof UnownedMetadataLineAddress ) {
         UnownedMetadataLineAddress other = (UnownedMetadataLineAddress) o;
         return this.lineAddr == other.lineAddr;
      } else if ( o instanceof MetadataLineAddress ) {
         // NB: this comparison with (owned) MetadataLineAddresses arises in cache lookups
         MetadataLineAddress other = (MetadataLineAddress) o;
         return this.lineAddr == other.lineAddr;
      }
      return false;
   }

   @Override
   public String toString() {
      return "umla:" + String.format("%d", this.lineAddr);
   }
}

/** The address of a metadata byte. */
class MetadataByteAddress extends ByteAddress {

   protected final CpuId owner;

   /**
    * @param a
    *           data address (it gets translated appropriately)
    * @param owner
    *           the cpu whose metadata this is
    */
   public MetadataByteAddress(long a, CpuId owner) {
      // this is a 2:1 data:metadata mapping
      super(a * MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
      assert owner != null;
      this.owner = owner;
      // metadata addresses should be aligned to METADATA_BYTES_PER_DATA_BYTE boundaries
      assert 0 == (addr % MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
   }

   /**
    * Return the location of the metadata corresponding to the given data address.
    * 
    * @param dab
    *           data address (it gets translated appropriately)
    * @param owner
    *           the cpu whose metadata this is
    */
   public MetadataByteAddress(DataByteAddress dab, CpuId owner) {
      // this is a 2:1 data:metadata mapping
      super(dab.get() * MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
      assert owner != null;
      this.owner = owner;
      // metadata addresses should be aligned to METADATA_BYTES_PER_DATA_BYTE boundaries
      assert 0 == (addr % MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
   }

   public MetadataByteAddress(MetadataLineAddress a) {
      // NB: no scaling
      super(a.get());
      assert a.owner() != null;
      this.owner = a.owner();
      // metadata addresses should be aligned to METADATA_BYTES_PER_DATA_BYTE boundaries
      assert 0 == (addr % MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
   }

   /** Copy constructor, but change ownership. */
   public MetadataByteAddress(MetadataByteAddress a, CpuId newOwner) {
      // NB: no scaling
      super(a.get());
      assert newOwner != null;
      this.owner = newOwner;
      // metadata addresses should be aligned to METADATA_BYTES_PER_DATA_BYTE boundaries
      assert 0 == (addr % MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
   }

   /** Copy constructor. */
   public MetadataByteAddress(MetadataByteAddress a) {
      // NB: no scaling
      super(a.get());
      assert a.owner() != null;
      this.owner = a.owner();
      // metadata addresses should be aligned to METADATA_BYTES_PER_DATA_BYTE boundaries
      assert 0 == (addr % MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
   }

   public static MetadataByteAddress Unscaled(long a, CpuId cpu) {
      a /= MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE();
      MetadataByteAddress mba = new MetadataByteAddress(a, cpu);
      return mba;
   }

   public UnownedMetadataByteAddress getUnowned() {
      return new UnownedMetadataByteAddress(this);
   }

   public CpuId owner() {
      return this.owner;
   }

   @Override
   public void incr() {
      this.addr += MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE();
   }

   @Override
   public void incr(long a) {
      this.addr += a;
      // metadata addresses should be aligned to METADATA_BYTES_PER_DATA_BYTE boundaries
      assert 0 == (addr % MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
   }

   @Override
   public MetadataLineAddress lineAddress() {
      long a = this.addr & (~MemorySystemConstants.LINE_OFFSET_MASK());
      return new MetadataLineAddress(a, owner);
   }

   /**
    * Line offset given in terms of metadata elements.
    */
   public short metadataLineOffset(RadishLine.Compacted c) {
      // NB: we always want to divide by the expanded size, since md has a fixed mapping (sized for
      // the expanded form). But the mask must change so we can index all the compact md elements.
      switch ( c ) {
      case COMPACT:
         short loffset = (short) (this.addr & MemorySystemConstants.COMPACT_MD_LINE_OFFSET_MASK());
         return (short) (loffset / MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
      case EXPANDED:
         return (short) (lineOffset() / MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
      }

      throw new IllegalArgumentException(c.toString());
   }

   @Override
   public String toString() {
      return "mba:" + String.format("%d", this.addr) + "," + this.owner.toString();
   }
}

/**
 * The address of a metadata byte, without an owner. Used when searching for metadata in an
 * owner-agnostic way.
 */
class UnownedMetadataByteAddress extends ByteAddress {

   /**
    * @param a
    *           owned metadata address (it does not undergo translation)
    */
   public UnownedMetadataByteAddress(MetadataByteAddress a) {
      super(a.get());
      // metadata addresses should be aligned to METADATA_BYTES_PER_DATA_BYTE boundaries
      assert 0 == (this.addr % MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
   }

   /**
    * @param a
    *           unowned metadata address (it does not undergo translation)
    */
   public UnownedMetadataByteAddress(UnownedMetadataByteAddress a) {
      super(a.get());
      // metadata addresses should be aligned to METADATA_BYTES_PER_DATA_BYTE boundaries
      assert 0 == (this.addr % MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
   }

   /**
    * NB: used by tests.
    * 
    * @param a
    *           data byte address that undergoes translation
    */
   UnownedMetadataByteAddress(long a) {
      // this is a 2:1 data:metadata mapping
      super(a * MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
      // metadata addresses should be aligned to METADATA_BYTES_PER_DATA_BYTE boundaries
      assert 0 == (this.addr % MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
   }

   @Override
   public void incr() {
      this.addr += MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE();
   }

   @Override
   public void incr(long a) {
      this.addr += a;
      // metadata addresses should be aligned to METADATA_BYTES_PER_DATA_BYTE boundaries
      assert 0 == (addr % MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
   }

   @Override
   public UnownedMetadataLineAddress lineAddress() {
      long a = this.addr & (~MemorySystemConstants.LINE_OFFSET_MASK());
      return new UnownedMetadataLineAddress(a);
   }

   /**
    * Line offset given in terms of metadata elements.
    */
   public short metadataLineOffset(RadishLine.Compacted c) {
      // NB: we always want to divide by the expanded size, since md has a fixed mapping (sized for
      // the expanded form). But the mask must change so we can index all the compact md elements.
      switch ( c ) {
      case COMPACT:
         short loffset = (short) (this.addr & MemorySystemConstants.COMPACT_MD_LINE_OFFSET_MASK());
         return (short) (loffset / MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
      case EXPANDED:
         return (short) (lineOffset() / MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE());
      }

      throw new IllegalArgumentException(c.toString());
   }

   @Override
   public boolean equals(Object o) {
      if ( o instanceof UnownedMetadataByteAddress ) {
         UnownedMetadataByteAddress other = (UnownedMetadataByteAddress) o;
         return this.addr == other.addr;
      }
      return false;
   }

   @Override
   public int hashCode() {
      return (int) this.addr;
   }

   @Override
   public String toString() {
      return "umba:" + String.format("%d", this.addr);
   }
}

/** A Radish epoch value, typically constructed from EpochBase + EpochOffset */
class Epoch implements Comparable<Epoch> {
   private final boolean                 touched;
   private final long                    value;

   /** Represents the empty (untouched) epoch, before any synchronization has occurred. */
   public static final Epoch             EMPTY_EPOCH = new Epoch(0, false);

   public static final Comparison<Epoch> Compare     = new Comparison<Epoch>();

   /** @return whether a happens-after b */
   public static boolean ha(Epoch a, Epoch b) {
      return Epoch.Compare.gte(a, b);
   }

   /** @return whether a happens-before b */
   public static boolean hb(Epoch a, Epoch b) {
      return Epoch.Compare.lt(a, b);
   }

   public Epoch(long v, boolean t) {
      value = v;
      touched = t;
   }

   public long get() {
      return this.touched ? value : 0;
   }

   public boolean touched() {
      return touched;
   }

   public boolean untouched() {
      return !touched;
   }

   public EpochOffset subtract(EpochBase b) {
      return new EpochOffset(this.value - b.get());
   }

   @Override
   public boolean equals(Object o) {
      if ( o instanceof Epoch ) {
         Epoch other = (Epoch) o;
         return this.touched == other.touched && this.value == other.value;
      }
      return false;
   }

   @Override
   public int hashCode() {
      return (int) this.value;
   }

   @Override
   public int compareTo(Epoch e) {
      if ( this.untouched() && e.untouched() ) {
         // throw new IllegalArgumentException("Comparing two untouched Epochs");
         return 0;
      } else if ( this.untouched() && e.touched() ) {
         return -1;
      } else if ( this.touched() && e.untouched() ) { return 1; }
      // two valid Epochs
      if ( this.value < e.value ) return -1;
      else if ( this.value == e.value ) return 0;
      else return 1;
   }

   @Override
   public String toString() {
      return "e:" + (this.touched ? this.value : "_");
   }

} // end class Epoch

/** The base value of an epoch, stored at each processor */
class EpochBase {
   public static final EpochBase INITIAL_EPOCH_BASE = new EpochBase(0);

   private final long            base;

   public EpochBase(long b) {
      this.base = b;
   }

   public Epoch add(EpochOffset o) {
      return new Epoch(this.base + o.get(), o.touched());
   }

   @Override
   public boolean equals(Object o) {
      if ( o instanceof EpochBase ) {
         EpochBase other = (EpochBase) o;
         return this.base == other.base;
      }
      return false;
   }

   @Override
   public int hashCode() {
      return (int) this.base;
   }

   @Override
   public String toString() {
      return "eb:" + this.base;
   }

   /**
    * NB: this should only ever be called by Epoch.subtract()
    * 
    * @deprecated marking as deprecated to discourage its use elsewhere
    */
   long get() {
      return this.base;
   }
}

/** The epoch offset, stored in in-cache metadata. This class is immutable. */
class EpochOffset implements Comparable<EpochOffset> {

   private final boolean                       touched;
   private final long                          offset;

   /** Represents the fact that the local processor hasn't accessed this byte. */
   public static final EpochOffset             UNTOUCHED_EPOCH_OFFSET = new EpochOffset();
   /** Represents the first epoch offset, in which a thread starts executing. */
   public static final EpochOffset             INITIAL_EPOCH_OFFSET   = new EpochOffset(1);

   public static final Comparison<EpochOffset> Compare                = new Comparison<EpochOffset>();

   public EpochOffset() {
      touched = false;
      offset = 0;
   }

   public EpochOffset(long o) {
      this.touched = true;
      this.offset = o;
   }

   /** Allocates a new EpochOffset, with a value 1 greater than this EpochOffset. */
   public EpochOffset getIncrementedEpochOffset() {
      return new EpochOffset(this.offset + 1);
   }

   /**
    * NB: this should only ever be called by EpochBase.add()
    * 
    * @deprecated marking as deprecated to discourage its use elsewhere
    */
   long get() {
      return offset;
   }

   public boolean touched() {
      return touched;
   }

   public boolean untouched() {
      return !touched;
   }

   @Override
   public boolean equals(Object o) {
      if ( o instanceof EpochOffset ) {
         EpochOffset other = (EpochOffset) o;
         return this.touched == other.touched && this.offset == other.offset;
      }
      return false;
   }

   @Override
   public int hashCode() {
      return (int) this.offset;
   }

   @Override
   public int compareTo(EpochOffset e) {
      if ( this.untouched() && e.untouched() ) {
         // throw new IllegalArgumentException("Comparing two untouched EpochOffsets");
         return 0;
      } else if ( this.untouched() && e.touched() ) {
         return -1;
      } else if ( this.touched() && e.untouched() ) { return 1; }
      // two valid EpochOffsets
      if ( this.offset < e.offset ) return -1;
      else if ( this.offset == e.offset ) return 0;
      else return 1;
   }

   @Override
   public String toString() {
      return "eo:" + (this.touched ? this.offset : "_");
   }

} // end class EpochOffset
