package edu.washington.cs.radishsim;

import static org.junit.Assert.*;

import java.util.Random;

import org.junit.Before;
import org.junit.Test;

public class AsyncCheckerCoreTests {

   static Machine<RadishLine>            m;

   static final CpuId                    P0                   = new CpuId(0);
   static final CpuId                    P1                   = new CpuId(1);
   static final CpuId                    CP0                  = new CpuId(2);
   static final int                      CORES                = 2;
   static final int                      L1_ASSOC             = 2;
   static final int                      LINE_SIZE            = 4;
   static final int                      L1_CACHE_SIZE        = 16;          // 4 blocks, 2 sets
   static final int                      BITS_PER_LOCAL_EPOCH = 2;

   static {
      MemorySystemConstants.unsafeSetLineSize(LINE_SIZE);
      assert RadishSim.XASSERTS;
   }

   static CacheConfiguration<RadishLine> l1config             = new CacheConfiguration<RadishLine>() {
                                                                 {
                                                                    cacheSize = L1_CACHE_SIZE;
                                                                    lineSize = LINE_SIZE;
                                                                    assoc = L1_ASSOC;
                                                                    level = CacheLevel.L1;
                                                                 }
                                                              };

   @Before
   public void setUp() {
      Machine.MachineParams<RadishLine> params = new DefaultTestingMachineParams() {

         @Override
         int numProcessors() {
            return CORES;
         }

         @Override
         CacheConfiguration<RadishLine> l1config() {
            return l1config;
         }

         @Override
         int bitsPerLocalEpochOffset() {
            return BITS_PER_LOCAL_EPOCH;
         }

         @Override
         boolean simulateSoftwareChecks() {
            return true;
         }

         @Override
         int asyncProcessingCores() {
            return 1;
         }
      };

      m = new Machine<RadishLine>(params);
   }

   // jargon:
   // Ad means data at location A
   // Amd=B means the metadata for data location A resides at location B

   @Test
   public void test_multiproSimple() {
      Processor<RadishLine> p = m.getProc(P0);
      CheckerProcessor<RadishLine> cp = m.getCheckerProc(CP0);

      // bring 0d,0md=0 into same set
      m.cacheRead(P0, 0, 1);
      assertEquals(1, p.stats.pc_l1d.pc_ReadMisses.get()); // data read
      assertEquals(1, p.stats.pc_l1md.pc_ReadMisses.get()); // md read
      assertEquals(0, p.stats.pc_MetadataEvictions.get());

      assertEquals(1, p.stats.pc_SwVcChecksTotal.get());
      cp.flushChecks(P0);
      // lock acquire and sw vc init execute on checker core
      assertEquals(2, cp.stats.pc_l1d.pc_WriteMisses.get());
   }

   @Test
   public void test_randomAccesses() {
      Random rng = new Random(4321);

      for ( int i = 0; i < 50000; i++ ) {
         int addr = rng.nextInt() % 256;
         boolean proc = rng.nextBoolean();
         CpuId cpu = proc ? P0 : P1;
         boolean read = rng.nextBoolean();

         boolean doSyncOp = (rng.nextInt() % 10) == 0;
         if ( doSyncOp )
            m.syncOp(cpu, read ? SynchronizationOperationType.SOURCE
                              : SynchronizationOperationType.SOURCE, false, addr);

         int size = rng.nextBoolean() ? 1 : 2;
         if ( read ) m.cacheRead(cpu, addr, size);
         else m.cacheWrite(cpu, addr, size);
      }
   }

}
