package edu.washington.cs.radishsim;

import static org.junit.Assert.*;

import java.util.Random;

import org.junit.Before;
import org.junit.Test;

import edu.washington.cs.radishsim.Machine.SimulationMode;

public class FastTrackEmulationTests {

   static Machine<RadishLine>            m;

   static final CpuId                    P0                   = new CpuId(0);
   static final CpuId                    P1                   = new CpuId(1);
   static final int                      CORES                = 2;
   static final int                      L1_ASSOC             = 2;
   static final int                      LINE_SIZE            = 4;
   static final int                      L1_CACHE_SIZE        = 16;          // 4 blocks, 2 sets
   static final int                      BITS_PER_LOCAL_EPOCH = 2;

   static {
      MemorySystemConstants.unsafeSetLineSize(LINE_SIZE);
      assert RadishSim.XASSERTS;
   }

   static CacheConfiguration<RadishLine> l1config             = new CacheConfiguration<RadishLine>() {
                                                                 {
                                                                    cacheSize = L1_CACHE_SIZE;
                                                                    lineSize = LINE_SIZE;
                                                                    assoc = L1_ASSOC;
                                                                    level = CacheLevel.L1;
                                                                 }
                                                              };

   @Before
   public void setUp() {
      Machine.MachineParams<RadishLine> params = new DefaultTestingMachineParams() {

         @Override
         SimulationMode simulationMode() {
            return SimulationMode.FASTTRACK;
         }
         
         @Override
         int numProcessors() {
            return CORES;
         }

         @Override
         CacheConfiguration<RadishLine> l1config() {
            return l1config;
         }

         @Override
         int bitsPerLocalEpochOffset() {
            return BITS_PER_LOCAL_EPOCH;
         }
         
         @Override
         boolean simulateSoftwareChecks() {
            return true;
         }
      };

      m = new Machine<RadishLine>(params);
   }

   @Test
   public void test_multiproSimple() {
      Processor<RadishLine> p = m.getProc(P0);

      m.cacheRead(P0, 0, 1);
      assertEquals(1, p.stats.pc_l1d.pc_ReadMisses.get()); // data read
      
      assertEquals(0, p.stats.pc_l1md.pc_ReadMisses.get()); // no md for FastTrack
      assertEquals(0, p.stats.pc_MetadataEvictions.get());
      assertEquals(0, p.stats.pc_SwVcChecksTotal.get());
      
      // lock acquire and vc init
      assertEquals(2, p.stats.pc_l1d.pc_WriteMisses.get());

//      // fetch 16d,16md=32, which map to set 0 and evict 0d,0md=0
//      m.cacheRead(P0, L1_CACHE_SIZE, 1);
//      assertEquals(2, p.stats.pc_l1md.pc_ReadMisses.get());
//      assertEquals(1, p.stats.pc_MetadataEvictions.get());
//
//      // fetch 0d,0md=0 again, evicting 16d,16md=32
//      m.cacheRead(P0, 0, 1);
//      assertEquals(3, p.stats.pc_l1md.pc_ReadMisses.get());
//      assertEquals(2, p.stats.pc_MetadataEvictions.get());
   }
   
   @Test
   public void test_randomAccesses() {
      Random rng = new Random(4321);

      for ( int i = 0; i < 50000; i++ ) {
         int addr = rng.nextInt() % 256;
         boolean proc = rng.nextBoolean();
         CpuId cpu = proc ? P0 : P1;
         boolean read = rng.nextBoolean();

         boolean doSyncOp = (rng.nextInt() % 10) == 0;
         if ( doSyncOp )
            m.syncOp(cpu, read ? SynchronizationOperationType.SOURCE
                               : SynchronizationOperationType.SOURCE, false, addr);

         int size = rng.nextBoolean() ? 1 : 2;
         if ( read ) m.cacheRead(cpu, addr, size);
         else m.cacheWrite(cpu, addr, size);
      }
   }

}
