package edu.washington.cs.radishsim;

import static org.junit.Assert.*;

import java.util.Random;

import org.junit.Before;
import org.junit.Test;

public class TypeSafetyOptTests {
   static Machine<RadishLine>            m;

   static final CpuId                    P0                   = new CpuId(0);
   static final CpuId                    P1                   = new CpuId(1);
   static final int                      CORES                = 2;
   static final int                      LINE_SIZE            = 4;
   static final int                      L1_ASSOC             = 2;
   static final int                      L1_CACHE_SIZE        = 16;          // 4 lines, 2 sets
   static final int                      L3_ASSOC             = 2;
   static final int                      L3_CACHE_SIZE        = 8;           // 2 lines, 1 set
   // (fully associative)
   static final int                      BITS_PER_LOCAL_EPOCH = 2;

   static {
      MemorySystemConstants.unsafeSetLineSize(LINE_SIZE);
      assert RadishSim.XASSERTS;
   }

   static CacheConfiguration<RadishLine> l1config             = new CacheConfiguration<RadishLine>() {
                                                                 {
                                                                    lineSize = LINE_SIZE;
                                                                    cacheSize = L1_CACHE_SIZE;
                                                                    assoc = L1_ASSOC;
                                                                    level = CacheLevel.L1;
                                                                 }
                                                              };
   static CacheConfiguration<RadishLine> l3config             = new CacheConfiguration<RadishLine>() {
                                                                 {
                                                                    lineSize = LINE_SIZE;
                                                                    cacheSize = L3_CACHE_SIZE;
                                                                    assoc = L3_ASSOC;
                                                                    level = CacheLevel.L3;
                                                                 }
                                                              };

   @Before
   public void setUp() {
      Machine.MachineParams<RadishLine> params = new DefaultTestingMachineParams() {

         @Override
         int numProcessors() {
            return CORES;
         }

         @Override
         int bitsPerLocalEpochOffset() {
            return BITS_PER_LOCAL_EPOCH;
         }

         @Override
         CacheConfiguration<RadishLine> l1config() {
            return l1config;
         }

         @Override
         boolean useL3() {
            return true;
         }

         @Override
         CacheConfiguration<RadishLine> l3config() {
            return l3config;
         }

         @Override
         boolean useClockGC() {
            return true;
         }

         @Override
         boolean useTypeSafetyOpt() {
            return true;
         }

      };
      m = new Machine<RadishLine>(params);
   }

   // jargon:
   // Ad means data at location A
   // Amd=B means the metadata for data location A resides at location B

   @Test
   public void test_evictLine() {
      Processor<RadishLine> p0 = m.getProc(P0);

      // 2-byte access sets epochs but doesn't allocate anything
      m.cacheRead(P0, 0, 2);
      assertEquals(1, p0.stats.pc_l3md.pc_ReadMisses.get());
      assertEquals(0, p0.stats.pc_MetadataEvictions.get());

      /*
       * Evict 0md=0. All d and md addrs map to set 0, so we only need half the associativity to
       * kick 0md=0 out.
       */
      for ( int k = 1; k <= (L1_ASSOC + L3_ASSOC) / 2; k++ ) {
         m.cacheRead(P0, k * L1_CACHE_SIZE, 1);
      }
      assertEquals(1, p0.stats.pc_MetadataEvictions.get());
   }

   @Test
   public void test_multiByteAccessDoesntAllocate() {
      Processor<RadishLine> p0 = m.getProc(P0);

      for ( int i = 0; i < (1 << (BITS_PER_LOCAL_EPOCH * 4)); i++ ) {
         // 2-byte access sets epochs but doesn't allocate anything
         m.cacheRead(P0, 0, 2);
         assertEquals(0, p0.stats.pc_EpochRollovers.get());
         m.syncOp(P0, SynchronizationOperationType.SOURCE, false, 0x42);
      }
   }

   @Test
   public void test_randomAccesses() {
      Random rng = new Random(123);

      for ( int i = 0; i < 50000; i++ ) {
         int addr = rng.nextInt() % 256;
         boolean proc = rng.nextBoolean();
         CpuId cpu = proc ? P0 : P1;
         boolean read = rng.nextBoolean();

         boolean doSyncOp = (rng.nextInt() % 10) == 0;
         if ( doSyncOp )
            m.syncOp(cpu, read ? SynchronizationOperationType.SOURCE
                              : SynchronizationOperationType.SOURCE, false, addr);

         int size = rng.nextBoolean() ? 1 : 2;
         if ( read ) m.cacheRead(cpu, addr, size);
         else m.cacheWrite(cpu, addr, size);
      }
   }

}
